from .engine import XPlant, NodeBase


def NodeDescriptor(*a):
    pass


class PYamlNode(NodeBase):
    _single_indent = '  '
    __slots__ = "value",

    def __init__(self, plant_instance, value):
        super(PYamlNode, self).__init__(plant_instance)
        self.value = value

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.value)

    def string_items(self, tree_level):
        representation = "{} - {}".format(self._single_indent * tree_level, self.value)
        if self.children:
            representation += ":"
        yield representation + "\n"

        for child in self.children:
            for item in child.string_items(tree_level + 1):
                yield item

    def node(self, value):
        return self._make_child(PYamlNode, value)

    def leaf(self, value):
        return self._make_child(PYamlNode, value)


class PYamlPlant(XPlant):
    _root_node_type = PYamlNode



from xplant.engine import LeafNodeMixin
from .engine import XPlant, NodeBase
from ._xml_attributes import XmlAttributes


class _XmlMarkupBase(NodeBase):
    __slots__ = "_value",
    _format = "{}"

    def __init__(self, plant_instance, value):
        super(_XmlMarkupBase, self).__init__(plant_instance)
        self._value = value

    def string_items(self, tree_level):
        yield self._format.format(self._value)


class Text(LeafNodeMixin, _XmlMarkupBase):
    pass


class Comment(LeafNodeMixin, _XmlMarkupBase):
    _format = "<!-- {} -->"


class CdataTag(LeafNodeMixin, _XmlMarkupBase):
    _format = "<![CDATA[{}]]>"


class _XmlElementNodeBase(NodeBase):
    attribute_processor = XmlAttributes
    _force_inline = False
    __slots__ = ("tag_name", "attributes",)
    indent_value = 2 * " "

    def __init__(self, plant_instance, tag_name, *args, **kwargs):
        super(_XmlElementNodeBase, self).__init__(plant_instance)
        self.tag_name = tag_name
        self.attributes = self.attribute_processor(*args, **kwargs)

    @classmethod
    def _break_line(cls, tree_level):
        indent = cls.indent_value
        return "\n" + indent * tree_level


class EmptyXmlElement(LeafNodeMixin, _XmlElementNodeBase):
    """ E.g.: <br />, <hr /> or <img src="#" alt="" /> """

    def string_items(self, tree_level):
        yield "<{}{} />".format(self.tag_name, self.attributes)


class XmlElement(_XmlElementNodeBase):

    def node(self, name, *p, **k):
        p = (name,) + p
        return self._make_child(XmlElement, *p, **k)

    def leaf(self, name, *p, **k):
        p = (name,) + p
        return self._make_child(EmptyXmlElement, *p, **k)

    def text(self, string_value):
        return self._make_child(Text, string_value)

    def cdata(self, string_value):
        return self._make_child(CdataTag, string_value)

    def comment(self, string_value):
        return self._make_child(Comment, string_value)

    def line(self, tag_name, content, *args, **kwargs):
        """ Element containing just a text. """
        with self.node(tag_name, *args, **kwargs):
            self.text(content)

    def string_items(self, tree_level):
        yield "<{}{}>".format(self.tag_name, self.attributes)
        for item in self._children_markup(tree_level):
            yield item
        yield "</{}>".format(self.tag_name)

    def _children_markup(self, tree_level):
        break_lines = self.degree > 1 and not self._force_inline
        child_indent = self._break_line(tree_level + 1)

        for child in self.children:
            if break_lines:
                yield child_indent

            for child_string in child.string_items(tree_level + 1):
                yield child_string

        if break_lines:
            yield self._break_line(tree_level)


class XmlPlant(XPlant):
    _root_node_type = XmlElement



import weakref


class NodeBase(object):
    __slots__ = ("children", "__plant")

    def __init__(self, plant_instance):
        self.children = []
        self.__plant = weakref.ref(plant_instance) if plant_instance else None

    @property
    def plant(self):
        return self.__plant() if self.__plant else None

    def __enter__(self):
        plant = self.plant
        if plant:
            plant.enter_(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        plant = self.plant
        if plant:
            plant.exit_()

    def __str__(self):
        return "".join(item for item in self.string_items(0) if item)

    def string_items(self, tree_level):
        msg = "\n".join([
            "NodeBase.string_items is an abstract method.", "",
            "It's supposed to:",
            " - be a generator yielding string items,",
            " - compose document representation of that node with ''.join(node.string_items(0)),",
            " - care for document's line breaks, markup and indentation.",
            "It needs to be implemented in derived class {name!r}.",
            ""
        ])
        raise NotImplementedError(msg.format(name=self.__class__.__name__))

    def append_child_node(self, new_node_obj):
        if not isinstance(new_node_obj, NodeBase):
            raise TypeError("Can append only instances of NodeBase, got {}.".format(type(new_node_obj).__name__))
        self.children.append(new_node_obj)

    @property
    def degree(self):
        """ For a given node, its number of children. A leaf is necessarily degree zero. """
        return len(self.children)

    @property
    def height(self):
        """ The height of a node is the number of edges on the longest path between that node and a leaf. """
        if not self.children:
            return 0
        else:
            return 1 + max(child.height for child in self.children)

    def _make_child(self, node_class, *args, **kwargs):
        if not issubclass(node_class, NodeBase):
            raise TypeError("Node class has to be a subclass of NodeBase, got %s." % node_class.__name__)
        new_node_instance = node_class(self.plant, *args, **kwargs)
        self.append_child_node(new_node_instance)
        return new_node_instance

    @classmethod
    def make_root(cls, plant_instance):
        """ Modify own class to be able to create a document plant - i.e. forest - collection of disjoint trees.
            IT can appear only on document's zero level, """

        class RootNode(cls):
            def __init__(self):
                NodeBase.__init__(self, plant_instance)

            def string_items(self, _):
                for child in self.children:
                    for item in child.string_items(tree_level=0):
                        yield item

        RootNode.__name__ = "{}Root".format(cls.__name__)
        return RootNode()


class LeafNodeMixin(object):
    """ Leaf - a node that is not allowed to have children. """

    def append_child_node(self, new_node_obj):
        raise AttributeError("{} is not allowed to append any children".format(self.__class__.__name__))

    def __enter__(self):
        msg = "Instance of {} cannot enter its scope, it's a leaf. Call it as a regular method (w/o 'with')."
        raise AttributeError(msg.format(self.__class__.__name__))


class XPlant(object):
    _root_node_type = None

    def __init__(self):
        assert self._root_node_type, "Undefined primary node for {}".format(self.__class__.__name__)
        assert issubclass(self._root_node_type, NodeBase), "Wrong primary node type {}".format(
            self._root_node_type)

        self.__tag_stack = [self._root_node_type.make_root(self)]

    def __str__(self):
        return str(self.__tag_stack[0])

    def __repr__(self):
        return "{}(<{}>)".format(type(self).__name__, type(self.__tag_stack[0]).__name__)

    @property
    def _top_node(self):
        """ The node whose context we are currently in. """
        return self.__tag_stack[-1]

    def __getattr__(self, attribute_name):
        """ Expose interface of a node that is currently on top (whose context we are currently in). """
        is_not_protected = not attribute_name.startswith("_")
        if is_not_protected and hasattr(self._top_node, attribute_name):
            return object.__getattribute__(self._top_node, attribute_name)

        return object.__getattribute__(self, attribute_name)

    def enter_(self, obj):
        self.__tag_stack.append(obj)

    def exit_(self):
        self.__tag_stack.pop()



from setuptools import setup, find_packages

long_description = open('README.rst').read()

setup(
    name='xplant',
    version='0.2',
    author='Michal Kaczmarczyk',
    author_email='michal.s.kaczmarczyk@gmail.com',
    maintainer='Michal Kaczmarczyk',
    maintainer_email='michal.s.kaczmarczyk@gmail.com',
    license='MIT license',
    url='https://gitlab.com/kamichal/xplant',
    description='Tree model markup builder.',
    long_description=long_description,
    packages=find_packages(),
    requires=[],
    install_requires=[],
    keywords='pythonic xml builder generator markup yaml',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Telecommunications Industry',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'Topic :: Database :: Front-Ends',
        'Topic :: Documentation',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Browsers',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
    ]
)



from ._six import str_types


class XmlAttributes(dict):
    """
        # A bit different dictionary. Can be created out of single string arguments,
        # tuples with two elements or by keyword arguments.
        # Its __str__ method returns well formed, reproducible xml attributes representation.

        from xplant.attrs import XmlAttributes

        a = XmlAttributes()
        assert str(a) == ''
        assert a == {}

        b = XmlAttributes("one", ("two", 2), three=3)
        # note sorting below
        assert str(b) == ' one three="3" two="2"'
        assert b == {'one': None, 'three': 3, 'two': 2}
    """
    default_attribute_rank = 100
    attribute_precedence_rank = {}
    attribute_name_substitutes = {}
    attribute_value_substitutes = {
        # keys need be written as repr of actual key (because of dict/hash aliasing e.g. False with 0)
        "True": "true",
        "False": "false",
    }

    def __init__(self, *strings_or_tuples, **keyword_arguments):
        super(XmlAttributes, self).__init__(self._normalize_arguments(strings_or_tuples, keyword_arguments))

    def __str__(self):
        return ''.join(self._generate_xml_representation())

    def _generate_xml_representation(self):
        sorted_items = sorted(self.items(), key=self._attribute_name_sorting_key)
        for key, value in sorted_items:
            if value is None:
                yield " {}".format(key)
            else:
                normalized_value = self.attribute_value_substitutes.get(str(value), value)
                yield ' {}="{}"'.format(key, normalized_value)

    @classmethod
    def _attribute_name_sorting_key(cls, element):
        attribute_name = element[0]
        attribute_rank = cls.attribute_precedence_rank.get(attribute_name, cls.default_attribute_rank)
        return attribute_rank, attribute_name

    @classmethod
    def _normalize_arguments(cls, positional_arguments, keyword_arguments):
        for argument in positional_arguments:
            if isinstance(argument, tuple):
                if len(argument) != 2:
                    raise ValueError("Attribute argument must be tuple of 2 elements (name, value).")
                yield argument
            elif isinstance(argument, str_types):
                yield (argument, None)
            else:
                raise ValueError("Couldn't make an attribute & value pair out of {!r}.".format(argument))

        for keyword, keyword_value in keyword_arguments.items():
            reclaimed_keyword = cls.attribute_name_substitutes.get(keyword, keyword)
            yield reclaimed_keyword, keyword_value
